/*
!=====================================================================
!
!          S p e c f e m 3 D  G l o b e  V e r s i o n  7 . 0
!          --------------------------------------------------
!
!     Main historical authors: Dimitri Komatitsch and Jeroen Tromp
!                        Princeton University, USA
!                and CNRS / University of Marseille, France
!                 (there are currently many more authors!)
! (c) Princeton University and CNRS / University of Marseille, April 2014
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 3 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
!
!=====================================================================
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <string.h>

#ifdef WITH_MPI
#include <mpi.h>
#endif

#define MAX(a, b) (((a) > (b)) ? (a) : (b))

/************************************************************************

helper functions for debugging

please uncomment if needed...

*************************************************************************/


/*
void save_to_max_surface_file_ (float *maxval) {
  int rank;
  char filename[BUFSIZ];
  FILE *fp;
#ifdef WITH_MPI
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);
#else
  rank = 0;
#endif
  sprintf (filename, "maxval_surface_proc_%03d.dat", rank);
  fp = fopen (filename, "a+");
  if (fp == NULL) {
    printf ("FILE Error:%s\n", (char *) strerror (errno));
    perror ("file error\n");
    exit (1);
  }
  fprintf (fp, "%e\n", *maxval);
  fclose (fp);
}
*/

/*
void save_fvector_ (float *vector, int *size, int *id, int *cpu_or_gpu) {
  FILE *fp;
  char filename[BUFSIZ];

  if (*cpu_or_gpu == 0) {
    sprintf (filename, "debug_output_cpu_%d.dat", *id);
  } else {
    sprintf (filename, "debug_output_gpu_%d.dat", *id);
  }

  fp = fopen (filename, "wb");
  if (fp == NULL) {
    printf ("FILE Error:%s\n", (char *) strerror (errno));
    perror ("file error\n");
    exit (1);
  }
  printf ("writing vector, vector[0]=%e\n", vector[0]);
  fwrite (vector, sizeof (float), *size, fp);
  fclose (fp);
}
*/

/*
void save_ivector_ (int *vector, int *size, int *id, int *cpu_or_gpu) {
  FILE *fp;
  char filename[BUFSIZ];

  if (*cpu_or_gpu == 0) {
    sprintf (filename, "debug_output_cpu_%d.dat", *id);
  } else {
    sprintf (filename, "debug_output_gpu_%d.dat", *id);
  }

  fp = fopen (filename, "wb");
  if (fp == NULL) {
    printf ("FILE Error:%s\n", (char *) strerror (errno));
    perror ("file error\n");
    exit (1);
  }

  fwrite (vector, sizeof (int), *size, fp);
  fclose (fp);

}
*/


/*
void get_max_from_surface_file_ (int *nodes_per_iterationf, int *NSTEP) {
  int nodes_per_iteration = *nodes_per_iterationf;
  char filename[BUFSIZ];
  int procid;
  size_t res;

#ifdef WITH_MPI
  MPI_Comm_rank (MPI_COMM_WORLD, &procid);
#else
  procid = 0;
#endif
  sprintf (filename, "DATABASES_MPI/proc%06d_noise_surface.bin", procid);

  FILE *fp; int it;
  printf ("Opening %s for analysis\n", filename);
  fp = fopen (filename, "rb");
  if (fp == NULL) {
    printf ("FILE Error:%s\n", (char *) strerror (errno));
    perror ("file error\n");
    exit (1);
  }

  float *vector = (float *)malloc (nodes_per_iteration*sizeof (float));
  float max_val;
  int i;
  max_val = 0.0;
  for (it = 0; it < *NSTEP; it++) {
    int pos = (sizeof (float)*nodes_per_iteration) *(it);
    fseek (fp, pos, SEEK_SET);
    res = fread (vector, sizeof (float), nodes_per_iteration, fp);
    if (res != nodes_per_iteration) {printf ("Reading error:%s\n", (char *) strerror (errno)); exit (1);}
    for (i = 0; i < nodes_per_iteration; i++) {
      max_val = MAX (max_val, vector[i]);
    }
    if (it % 500 == 0) {
      printf ("scanning it=%d\n", it);
    }
  }
  printf ("max_val=%e\n", max_val);
}
*/

/*
void compare_two_vectors_exact_ (int *sizef, float *vector1, float *vector2, int *num_errors) {
  int size = *sizef;
  int i;
  int error_count = 0;

  for (i = 0; i < size; ++i) {
    if (vector1[i] != vector2[i]) {
      error_count++;
      if (error_count < 10) {
        printf ("err[%d]: %e != %e\n", i, vector1[i], vector2[i]);
      }
    }
  }
  printf ("** **Error Count: %d ****\n", error_count);
  *num_errors = error_count;
}
*/

/*
void compare_two_vectors_ (int *sizef, float *vector1, float *vector2, int *num_errors) {
  int size = *sizef;
  int i;
  int error_count = 0;

  for (i = 0; i < size; ++i) {
    if (vector1[i] != 0) {
      if (fabsf (vector1[i]-vector2[i])/vector1[i] > 0.01) {
        if (fabsf (vector1[i]-vector2[i]) > 1e-20) {
          error_count++;
          if (error_count<10) {
            printf ("err[%d]: %e != %e\n", i, vector1[i], vector2[i]);
          }
        }
      }
    }
  }
  printf ("** **Error Count: %d ****\n", error_count);
  *num_errors = error_count;
}
*/

/*
void compare_surface_files_ (int *bytes_per_iteration, int *number_of_iterations) {
  const char *cpu_file = "DATABASES_MPI/cpu_proc000001_noise_surface.bin";
  const char *gpu_file = "DATABASES_MPI/cpu_v2_proc000001_noise_surface.bin";
  size_t res;
  FILE *fp_cpu;

  fp_cpu = fopen (cpu_file, "rb");
  if (fp_cpu == NULL) {
    printf ("CPU FILE Error:%s\n", (char *) strerror (errno));
    perror ("CPU file error\n");
  }

  FILE *fp_gpu;
  fp_gpu = fopen (gpu_file, "rb");
  if (fp_gpu == NULL) {
    printf ("GPU FILE Error:%s\n", (char *) strerror (errno));
    perror ("GPU file error\n");
  }

  float *gpu_vector = (float *)malloc (*bytes_per_iteration);
  float *cpu_vector = (float *)malloc (*bytes_per_iteration);
  int i, it, error_count = 0;
  for (it = 0; it < *number_of_iterations; it++) {
    int pos = (*bytes_per_iteration) *(it);

    fseek (fp_cpu, pos, SEEK_SET);
    fseek (fp_gpu, pos, SEEK_SET);

    int number_of_nodes = *bytes_per_iteration / sizeof (float);
    res = fread (cpu_vector, sizeof (float), number_of_nodes, fp_cpu);
    if (res != number_of_nodes) {
      printf ("Reading error:%s\n", (char *) strerror (errno));
      exit (1);
    }

    res = fread (gpu_vector, sizeof (float), number_of_nodes, fp_gpu);
    if (res != number_of_nodes) {
      printf ("Reading error:%s\n", (char *) strerror (errno));
      exit (1);
    }

    int size = number_of_nodes;
    float gpu_min_val=10;
    float gpu_max_val = 0;
    float cpu_min_val=10;
    float cpu_max_val = 0;

    if (it < 100) {
      for (i = 0; i < size; i++) {
        if ((fabs(cpu_vector[i] - gpu_vector[i]) / (fabs(cpu_vector[i]) + 1e-31) > 0.01)) {
          if (error_count < 30)
            printf ("Error[%d]: %g != %g\n", i, cpu_vector[i], gpu_vector[i]);
          if (cpu_vector[i] > 1e-30)
            error_count++;
        }
        if (gpu_vector[i] > gpu_max_val) gpu_max_val = gpu_vector[i];
        if (gpu_vector[i] < gpu_min_val) gpu_min_val = gpu_vector[i];

        if (cpu_vector[i] > cpu_max_val) cpu_max_val = cpu_vector[i];
        if (cpu_vector[i] < cpu_min_val) cpu_min_val = cpu_vector[i];
      }
      printf ("%d Total Errors\n", error_count);
      printf ("size:%d\n", size);
      printf ("GPU:[min/max]=%e/%e\n", gpu_min_val, gpu_max_val);
      printf ("CPU:[min/max]=%e/%e\n", cpu_min_val, cpu_max_val);
    }
  }
  printf ("End of Surface Compare\n");
  exit (1);
}
*/

/*
void compare_fvector_ (float *vector, int *size, int *id, int *cpu_or_gpu) {
  FILE *fp;
  char cmp_filename[BUFSIZ];
  size_t res;

  float *compare_vector = (float *)malloc (*size*sizeof (float));
  if (*cpu_or_gpu == 0) {   //swap GPU/CPU to compare
    sprintf (cmp_filename, "debug_output_gpu_%d.dat", *id);
  }
  else {
    sprintf (cmp_filename, "debug_output_cpu_%d.dat", *id);
  }

  fp = fopen (cmp_filename, "rb");
  // read the values
  if (fp == NULL) {
    printf ("Cannot open comparison file %s.\n", cmp_filename);
    exit (1);
  }
  res = fread (compare_vector, sizeof (float), *size, fp);
  if (res != *size) {
    if (feof (fp))
      printf ("Premature end of file.");
    else
      printf ("File read error.");
  }

  fclose (fp);

  int i;
  int error_count = 0;
  for (i = 0;i<*size;i++) {
    if ((fabs(vector[i] - compare_vector[i]) / vector[i] > 0.0001)) {
      if (error_count < 30) {
        printf ("Error[%d]: %f != %f\n", i, compare_vector[i], vector[i]);
      }
      error_count++;
    }
  }
  printf ("%d Total Errors\n", error_count);
  printf ("size:%d\n", *size);
}
*/

/*
void compare_ivector_ (int *vector, int *size, int *id, int *cpu_or_gpu) {
  FILE *fp;
  char cmp_filename[BUFSIZ];
  size_t res;

  int *compare_vector = (int *) malloc (*size * sizeof (int));
  if (*cpu_or_gpu == 0) {   //swap GPU/CPU to compare
    sprintf (cmp_filename, "debug_output_gpu_%d.dat", *id);
  } else {
    sprintf (cmp_filename, "debug_output_cpu_%d.dat", *id);
  }

  // read the values
  fp = fopen (cmp_filename, "rb");
  if (fp == NULL) {
    printf ("Cannot open comparison file %s.\n", cmp_filename);
    exit (1);
  }

  res = fread (compare_vector, sizeof (int), *size, fp);
  if (res != *size) {
    if (feof (fp))
      printf ("Premature end of file.");
    else
      printf ("File read error.");
  }

  fclose (fp);

  int i;
  int error_count = 0;
  for (i = 0; i < *size; i++) {
    if ((fabs(vector[i] - compare_vector[i]) / vector[i] > 0.01) && error_count < 30) {
      printf ("Error[%d]: %d != %d\n", i, compare_vector[i], vector[i]);
      error_count++;
    }
  }
  printf ("%d Total Errors\n", error_count);
}
*/

